local _, class = UnitClass("player")
if class ~= "DEATHKNIGHT" then
	return
end

local plugin = mRunes:NewModule("Diseases", "AceEvent-3.0")
plugin.mRunesPlugin = true

local AceGUI = LibStub("AceGUI-3.0")
local LSM = LibStub("LibSharedMedia-3.0")
local BD = LibStub("LibBackdrop-1.0")
local UIF
local db

local LUnitDebuff = UnitDebuff

local InCombat = false

local Bars = {
}

local Spells = {
	FrostFever = {
		spell = 59921
	},
	BloodPlague = {
		spell = 59879
	},
	ScarletFever = {
		spell = 81130,
		talent = {
			tab = 1,
			index = 6,
		},
	},
}

-- local name, icon, tier, column, currentRank, maxRank = GetTalentInfo(tab, i, isnotplayer)

local backdrop = {
	tile = true, 
	tileSize = 16,
	insets = {},
}

plugin.DisplayName = "Diseases"
plugin.DisplayIcon = [[Interface\Icons\spell_deathknight_bloodplague]]
plugin.DisplayOrder = 3

function plugin:OnInitialize()
	local defaults = {
		profile = {
			Enabled = true,
			
			Attached = {
				["**"] = {
					x = 0,
					y = 0,
					Point = "TOP",
					RelPoint = "BOTTOM",
				}
			},			
			AttachedFrame = "Runic Power",
			
			Bar = {
				["**"] = {
					Enabled = true,
					DisableOnMissingTalent = true,
					RelativeHeight = 1,
					Color = {1, 1, 1, 1},
					Order = 1,
					ShowName = true,
					ShowTime = true,
					
					Font = "accid",
					FontSize = 16,
					FontFlags = "OUTLINE",
					FontColor = {1, 1, 1, 1},
					HorAlign = "LEFT",
					InternalSpacing = 2,
				},
			},
			
			Width = 330,
			Height = 40,							
			BarSpacing = 2,
			
			AlphaIncCombat = 1,
			AlphaOutCombat = 0.0,			
			
			Texture = "HalD",
			
			BorderTexture = "None",
			BorderWidth = 5,
			BorderInsets = 0,
			BorderPadding = 0,
			BorderColor = {0, 0, 0, 1},
			BackgroundColor = {0, 0, 0, 0.5},
		}
	}
	
	self.db = mRunes.db:RegisterNamespace(self:GetName(), defaults)
	self.db.RegisterCallback(self, "OnProfileChanged", "ProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "ProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "ProfileChanged")
	db = self.db.profile
	
	self:SetEnabledState(db.Enabled)
		
	UIF = LibStub("LibGUIFactory-1.0"):GetFactory("mRunes")
	
	if not self.DiseaseFrame then
		self:CreateDiseaseFrame()
	end
end

function plugin:OnEnable()	
	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_TALENT_UPDATE", "ApplySettings")
	mRunes.RegisterCallback(self, "EnteringCombat")
	mRunes.RegisterCallback(self, "LeavingCombat")
	
	self:UNIT_AURA("UNIT_AURA", "target")
	
	self:ApplySettings()
	
	if UnitExists("target") and  UnitReaction("player", "target") then
		mRunes.FadeIn(self.DiseaseFrame, false, db.AlphaIncCombat)
	else
		mRunes.FadeOut(self.DiseaseFrame, db.AlphaOutCombat)
	end
end

function plugin:OnDisable()
	mRunes:UnregisterOnUpdate(self.DiseaseFrame, self.Update)
	if self.DiseaseFrame then mRunes.FadeOut(self.DiseaseFrame) end
end

function plugin:ProfileChanged(event)
	db = self.db.profile
	self:ApplySettings()
end

-- ~~~~~~~~~~~~~~~~~~~~
-- Callbacks ~~~~~~~~~~
-- ~~~~~~~~~~~~~~~~~~~~
function plugin:EnteringCombat(event)
	if UnitExists("target") and  UnitReaction("player", "target") < 5 then
		mRunes.FadeIn(self.DiseaseFrame, false, db.AlphaIncCombat)
	end
	InCombat = true
end

function plugin:LeavingCombat(event)
	mRunes.FadeOut(self.DiseaseFrame, db.AlphaOutCombat)
	InCombat = false
end

function plugin:EnterWorld(...)
	if UnitExists("target") and  UnitReaction("player", "target") then
		mRunes.FadeIn(self.DiseaseFrame, false, db.AlphaIncCombat)
	else
		mRunes.FadeOut(self.DiseaseFrame, db.AlphaOutCombat)
	end
end

function plugin:Update()
	local t = GetTime()
	
	for name, bar in pairs(Bars) do
		local bdb = db.Bar[name]		
		if bdb.Enabled then
			if bar.duration then
				local dur = bar.expires - t
				bar:SetValue(dur / bar.duration)
				if bdb.ShowTime and bdb.ShowName then
					bar.text:SetFormattedText("%s: %.1f", bar.SpellName, dur)
				elseif bdb.ShowTime then
					bar.text:SetFormattedText("%.1f", dur)
				elseif bdb.ShowName then
					bar.text:SetFormattedText("%s", bar.SpellName)		
				else
					bar.text:SetText()
				end
			end
		end
	end
end

function plugin:GetAttachFrame()
	return self.DiseaseFrame
end

-- ~~~~~~~~~~~~~~~~~~~~
-- Events ~~~~~~~~~~~~~
-- ~~~~~~~~~~~~~~~~~~~~
function plugin:UNIT_AURA(event, unit)	
	local DiseaseFrame = self.DiseaseFrame
	if DiseaseFrame then
		if unit == "target" then
			DiseaseFrame.DoUpdate = false
			for i, bar in pairs(Bars) do
				local name, rank, icon, count, atype, duration, expires, caster, stealable, shouldConsolidate, spellId  = LUnitDebuff(unit, bar.SpellName)
				
				if name and caster == "player" then
					bar.expires = expires
					bar.duration = duration
					DiseaseFrame.DoUpdate = true
				else
					bar.expires = nil
					bar.duration = nil
					bar:SetValue(0)				
					if not db.AlwaysShowText then
						bar.text:SetText()
					else
						bar.text:SetText(0)
					end
				end
			end
			
			if DiseaseFrame.DoUpdate then
				mRunes:RegisterOnUpdate(self.DiseaseFrame, self.Update)
			else
				mRunes:UnregisterOnUpdate(self.DiseaseFrame, self.Update)
			end
		end
	end
end

function plugin:PLAYER_TARGET_CHANGED(event)
	if UnitExists("target") and  UnitReaction("player", "target") < 5 and UnitAffectingCombat("player") then
		self.DiseaseFrame:SetAlpha(db.AlphaIncCombat)
	else
		self.DiseaseFrame:SetAlpha(db.AlphaOutCombat)
	end
	self:UNIT_AURA("UNIT_AURA", "target")
end

-- ~~~~~~~~~~~~~~~~~~~~
-- Settings	~~~~~~~~~~~
-- ~~~~~~~~~~~~~~~~~~~~
local SortedBars = {}
local function SortBars(a, b)
	return a.db.Order < b.db.Order
end

function plugin:ApplySettings()
	local DiseaseFrame = self.DiseaseFrame
	if DiseaseFrame then
		local anchor = mRunes:GetAttachFrame(db.AttachedFrame) or mRunes.MainFrame
		DiseaseFrame:ClearAllPoints()
		DiseaseFrame:SetPoint(db.Attached[db.AttachedFrame].Point, anchor, db.Attached[db.AttachedFrame].RelPoint, db.Attached[db.AttachedFrame].x, db.Attached[db.AttachedFrame].y)
		DiseaseFrame:SetWidth(db.Width)
		DiseaseFrame:SetHeight(db.Height)
		
		backdrop.insets.left = db.BorderInsets
		backdrop.insets.right = db.BorderInsets
		backdrop.insets.top = db.BorderInsets
		backdrop.insets.bottom = db.BorderInsets
		backdrop.edgeFile = LSM:Fetch("border", db.BorderTexture)
		backdrop.edgeSize = db.BorderWidth
		backdrop.bgFile = LSM:Fetch("background", "Solid")
		DiseaseFrame:SetBackdrop(backdrop)
		DiseaseFrame:SetBackdropColor(unpack(db.BackgroundColor))
		DiseaseFrame:SetBackdropBorderColor(unpack(db.BorderColor))

		local BarFrame = DiseaseFrame.BarFrame
		BarFrame:ClearAllPoints()
		BarFrame:SetPoint("TOPLEFT", db.BorderPadding, -db.BorderPadding)
		BarFrame:SetPoint("TOPRIGHT", -db.BorderPadding, -db.BorderPadding)
		BarFrame:SetHeight(DiseaseFrame:GetHeight() - 2 * db.BorderPadding)
		local Height = BarFrame:GetHeight()
		local TotalRelative = 0
		
		wipe(SortedBars)
		for name, bar in pairs(Bars) do
			local bdb = db.Bar[name]
			bar.db = bdb
			
			if bdb.Enabled then
				local valid = true
				if bar.Restriction and bdb.DisableOnMissingTalent then
					if bar.Restriction.type == "TALENT" then
						local name, icon, tier, column, currentRank, maxRank = GetTalentInfo(bar.Restriction.tab, bar.Restriction.index, false)
						if not name or currentRank == 0 then
							valid = false
						end
					end
				end
				
				if valid then 
					TotalRelative = TotalRelative + bdb.RelativeHeight
					bar:Show()
					tinsert(SortedBars, bar)
				end
			else
				bar:ClearAllPoints()
				bar:Hide()
			end
		end
		table.sort(SortedBars, SortBars)
		
		for i, bar in pairs(SortedBars) do
			local bdb = bar.db
			local BarHeight = bdb.RelativeHeight * Height / TotalRelative - db.BarSpacing
			bar:ClearAllPoints()	
			if i == 1 then
				bar:SetPoint("TOPLEFT", 0, -db.BarSpacing / 2)
				bar:SetPoint("TOPRIGHT", 0, -db.BarSpacing / 2)
				bar:SetHeight(BarHeight)
			else
				bar:SetPoint("TOPLEFT", SortedBars[i-1], "BOTTOMLEFT", 0, - db.BarSpacing)
				bar:SetPoint("TOPRIGHT", SortedBars[i-1], "BOTTOMRIGHT",  0, - db.BarSpacing)
				bar:SetHeight(BarHeight)
			end					
			
			bar:SetStatusBarTexture(LSM:Fetch("statusbar", db.Texture))
			bar:SetStatusBarColor(unpack(bdb.Color))
			
			local leftinset, rightinset = 0, 0
			if bdb.HorAlign == "LEFT" then
				leftinset = bdb.InternalSpacing
			elseif bdb.HorAlign == "RIGHT" then
				rightinset = -bdb.InternalSpacing
			end	
				
			bar.text:SetPoint("TOPLEFT", leftinset, 0)
			bar.text:SetPoint("BOTTOMRIGHT", rightinset, 0)
			bar.text:SetFont(LSM:Fetch("font", bdb.Font), bdb.FontSize, bdb.FontFlags)
			bar.text:SetTextColor(unpack(bdb.FontColor))
			bar.text:SetJustifyH(bdb.HorAlign)
			bar.text:SetJustifyV("MIDDLE")
		end
		
		if InCombat and UnitExists("target") and UnitReaction("player", "target") < 5 then
			DiseaseFrame:SetAlpha(db.AlphaIncCombat)
		else
			DiseaseFrame:SetAlpha(db.AlphaOutCombat)
		end		
		if DiseaseFrame:GetAlpha() > 0 then DiseaseFrame:Show() else DiseaseFrame:Hide() end		
	end	
	
	if not db.Enabled and self:IsEnabled() then self:Disable()
	elseif db.Enabled and not self:IsEnabled() then self:Enable() end
end

-- ~~~~~~~~~~~~~~~~~~~~
-- Create Frames ~~~~~~
-- ~~~~~~~~~~~~~~~~~~~~
function plugin:CreateDiseaseFrame()
	local frame = CreateFrame("Frame", nil, UIParent)	
	BD:EnhanceBackdrop(frame)
	
	local barFrame = CreateFrame("Frame", nil, frame)
	frame.BarFrame = barFrame
	
	for s, data in pairs(Spells) do				
		local spellName, _, spellIcon = GetSpellInfo(data.spell)
		if spellName then
			local bar = self:CreateBar(barFrame, s)
			
			bar.Spell = data.spell
			bar.SpellName = spellName
			bar.SpellIcon = spellIcon
			
			if data.talent then
				bar.Restriction = {
					type = "TALENT",
					tab = data.talent.tab,
					index = data.talent.index,
				}
			end
			
			Bars[s] = bar
		end
	end
	
	self.DiseaseFrame = frame
end

function plugin:CreateBar(parent, BarName)
	local bar = CreateFrame("StatusBar", "mRunes"..BarName.."Bar", parent)
	bar:SetMinMaxValues(0,1)
	
	local text = bar:CreateFontString(nil, nil, "GameFontNormal")
	bar.text = text	
	
	local background = CreateFrame("StatusBar", "mRunes"..BarName.."BarBackground", bar)
	background:SetAllPoints(bar)
	bar.background = background
	
	bar:SetFrameLevel(background:GetFrameLevel() + 1)
	
	return bar
end

-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-- Options ~~~~~~~~~~~~~~~~~~~~
-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function plugin:GetOptions()
	return plugin.GetGUI
end

do
	local function Callback()
		plugin:ApplySettings()
	end
	
	local FontFlagList = {
		["NONE"] = "None",
		["OUTLINE"] = "Outline",
		["THICKOUTLINE"] = "Thick Outline",
		["MONOCHROME"] = "Monochrome",
	}
	
	local VerticalAlignList = {
		["TOP"] = "Top",
		["MIDDLE"] = "Middle",
		["BOTTOM"] = "Bottom",
	}
	
	local HorizontalAlignList = {
		["LEFT"] = "Left",
		["CENTER"] = "Center",
		["RIGHT"] = "Right",
	}

	local TabSelected
	local LastTabSelected
	local BarSelected
	local LastBarSelected
	local BarTabs = {}

	function plugin:GetGUI()
		self = plugin	
		
		local returnGroup = AceGUI:Create("SimpleGroup")
		returnGroup:SetFullWidth(true)
		returnGroup:SetLayout("Flow")
		
		local title = UIF:Title1(plugin.DisplayName)
		title:SetImage(plugin.DisplayIcon)
		returnGroup:AddChild(title)
		
		returnGroup:AddChild(UIF:NewLine(10))
		
		returnGroup:AddChild(UIF:CheckBox("Enabled", db, "Enabled", Callback, nil, 0.33))
		
		local tgroup = UIF:TabGroup()
		returnGroup:AddChild(tgroup)
		tgroup:SetTabs({
			{text = "Position & Size", value = "PaS"},
			{text = "Bars", value = "Bars"},
			{text = "Border & Alpha", value = "CaA"},		
		})		
		tgroup:SetCallback("OnGroupSelected", TabSelected)
		tgroup:SelectTab(LastSelectedTab or "PaS")
		
		return returnGroup
	end
	
	TabSelected = function(self, event, tab)
		self:ReleaseChildren()
		LastSelectedTab = tab
		
		if tab == "PaS" then
			self:AddChild(UIF:Text1("Select a frame to attach to. You can then adjust the relative positioning to that frame via X and Y as well as the anchor points."))			
			self:AddChild(UIF:Dropdown("Frame", db, "AttachedFrame", Callback, mRunes:GetAttachList(plugin), 0.5, true))
			self:AddChild(UIF:Spacer(10))
			self:AddChild(UIF:Text2("Avoid creating circle references!", nil, nil, nil, nil, nil, 0.4))
			self:AddChild(UIF:NewLine())			
			self:AddChild(UIF:Dropdown("Point", db.Attached[db.AttachedFrame], "Point", Callback, mRunes:GetAnchorList(), 0.5, true))
			self:AddChild(UIF:Dropdown("Relative Point", db.Attached[db.AttachedFrame], "RelPoint", Callback, mRunes:GetAnchorList(), 0.5, true))
			self:AddChild(UIF:Slider("X", db.Attached[db.AttachedFrame], "x", -300, 300, 0.1, Callback, 0.5))
			self:AddChild(UIF:Slider("Y", db.Attached[db.AttachedFrame], "y", -300, 300, 0.1, Callback, 0.5))
			self:AddChild(UIF:Slider("Width", db, "Width", 5, 1000, 0.1, Callback, 0.5))
			self:AddChild(UIF:Slider("Height", db, "Height", 5, 200, 0.1, Callback, 0.5))
			self:AddChild(UIF:Slider("Space between bars", db, "BarSpacing", 0, 20, 0.1, Callback, 0.5))
		
		elseif tab == "CaA" then
			local group = UIF:InlineGroup2("Alpha")			
			self:AddChild(group)
			group:AddChild(UIF:Slider("In combat", db, "AlphaIncCombat", 0, 1, 0.01, Callback, 0.33))
			group:AddChild(UIF:Slider("OOC or No target", db, "AlphaOutCombat", 0, 1, 0.01, Callback, 0.33))
			
			group = UIF:InlineGroup2("Border and Background")
			self:AddChild(group)
			group:AddChild(UIF:LSMDropdown("border", "Border texture", db, "BorderTexture", Callback, 0.5))
			group:AddChild(UIF:LSMDropdown("statusbar", "Bar Texture", db, "Texture", Callback, 0.5))				
			
			group:AddChild(UIF:ColorSelect("Background Color", db, "BackgroundColor", Callback, true, 0.5))			
			group:AddChild(UIF:ColorSelect("Border color", db, "BorderColor", Callback, true, 0.5))		
			group:AddChild(UIF:Slider("Border Width", db, "BorderWidth", 0, 50, 1, Callback, 0.33))
			group:AddChild(UIF:Slider("Border Padding", db, "BorderPadding", 0, 100, 0.1, Callback, 0.33))
			group:AddChild(UIF:Slider("Border Insets", db, "BorderInsets", 0, 50, 1, Callback, 0.33))	
			
		elseif tab == "Bars" then
			wipe(BarTabs)
			for name, bar in pairs(Bars) do
				tinsert(BarTabs, {text = bar.SpellName, value = name})
			end
			
			local tgroup = UIF:TabGroup()
				
			self:AddChild(tgroup)
			tgroup:SetTabs(BarTabs)		
			tgroup:SetCallback("OnGroupSelected", BarSelected)
			tgroup:SelectTab(LastBarSelected or BarTabs[1].value)
		end
		
		self:DoLayout()
		self.parent:DoLayout()
	end
	
	BarSelected = function(self, event, tab)
		self:ReleaseChildren()
		LastBarSelected = tab
		
		local bar = Bars[tab]
		local bdb = db.Bar[tab]
		
		self:AddChild(UIF:CheckBox("Enabled", bdb, "Enabled", Callback, nil, 0.3))
		self:AddChild(UIF:ColorSelect("Color", bdb, "Color", Callback, true, 0.5))		
		
		self:AddChild(UIF:NewLine())
		
		if bar.Restriction then
			if bar.Restriction.type == "TALENT" then
				self:AddChild(UIF:CheckBox("Disable when missing talent", bdb, "DisableOnMissingTalent", Callback, "This will disable this bar if the talent for it is missing", 0.8))
			end
			self:AddChild(UIF:NewLine())
		end		
		
		self:AddChild(UIF:CheckBox("Show Name", bdb, "ShowName", Callback, "Toggle display of spellname on the bar", 0.5))
		self:AddChild(UIF:CheckBox("Show Time", bdb, "ShowTime", Callback, "Toggle display of remaining time", 0.5))		
		self:AddChild(UIF:Slider("Order", bdb, "Order", 1, 5, 1, Callback, 0.5))		
		self:AddChild(UIF:Slider("Relative Height", bdb, "RelativeHeight", 1, 20, 1, Callback, 0.5))
		
		local group = UIF:InlineGroup2("Font")
		self:AddChild(group)
		group:AddChild(UIF:LSMDropdown("font", "Font", bdb, "Font", Callback, 0.5))
		group:AddChild(UIF:Dropdown("Font Flags", bdb, "FontFlags", Callback, FontFlagList, 0.5))		
		group:AddChild(UIF:ColorSelect("Color", bdb, "FontColor", Callback, false, 0.2))
		group:AddChild(UIF:Slider("Size", bdb, "FontSize", 2, 50, 1, Callback, 0.4))
		group:AddChild(UIF:Slider("Margin", bdb, "InternalSpacing", 0, 50, 0.1, Callback, 0.4))
		group:AddChild(UIF:Dropdown("Horizontal Align", bdb, "HorAlign", Callback, HorizontalAlignList, 0.5))
		
		self:DoLayout()
		self.parent:DoLayout()
	end
end